"""
clearindex.py
Ian Kollipara <ian.kollipara@gmail.com>

This module contains the ClearIndexCommand class for the Django MeiliSearch app.
"""

from django.apps import apps
from django.core.management.base import BaseCommand

from django_meili._client import client as _client
from django_meili.models import IndexMixin


class Command(BaseCommand):
    help = "Clears the MeiliSearch index for the given model."

    def add_arguments(self, parser):
        parser.add_argument("model", type=str, help="The model to clear the index for.")

    def handle(self, *args, **options):
        model = self._resolve_model(options["model"])
        index = _client.get_index(model._meilisearch["index_name"])
        task = index.delete_all_documents()
        finished = _client.wait_for_task(task.task_uid)
        if finished.status == "failed":
            raise Exception(finished)
        self.stdout.write(self.style.SUCCESS(f"Cleared index for {model}"))

    def _resolve_model(self, model: str) -> type[IndexMixin]:
        """
        Resolve the model from the given string.
        """
        try:
            Model = apps.get_model(model)
            if IndexMixin not in Model.__mro__:
                raise ValueError("Model does not inherit from IndexMixin")
        except LookupError as e:
            self.stdout.write(self.style.ERROR(f"Model not found: {model}"))
            exit(1)
        except ValueError as e:
            self.stdout.write(self.style.ERROR(f"Invalid model: {model}"))
            exit(1)
        return Model



"""
models.py
Ian Kollipara <ian.kollipara@gmail.com>

This module contains the models for the Django MeiliSearch app.
"""

from typing import Iterable, TypedDict

from django.conf import settings
from django.db import models
from meilisearch.models.task import TaskInfo

from ._client import client as _client
from .querysets import IndexQuerySet

# Create your models here.


class MeiliGeo(TypedDict):
    lat: float | str
    lng: float | str


class _Meili(TypedDict):
    primary_key: str
    index_name: str
    displayed_fields: Iterable[str] | None
    searchable_fields: Iterable[str] | None
    filterable_fields: Iterable[str] | None
    sortable_fields: Iterable[str] | None
    supports_geo: bool
    tasks: list[TaskInfo]


class IndexMixin(models.Model):
    """
    Mixin to provide Meilisearch Index for the given model.

    This mixin will create a Meilisearch index for the model and provide a
    queryset to interact with that index.

    To use this mixin, create a model that inherits from it and set the
    MeiliMeta class with the following attributes:
    - displayed_fields: The fields to display in search results.
    - searchable_fields: The fields to search on.
    - filterable_fields: The fields to filter on.
    - sortable_fields: The fields to sort on.
    - supports_geo: Whether the model supports geolocation.
    - index_name: The name of the index in Meilisearch.
    - primary_key: The primary key for the model.
    - include_pk_in_search: include the pk in the search results

    This mixin also defines a few methods that can be overridden:
    - meili_filter: A function to decide if the model should be added to meilisearch.
    - meili_serialize: How to serialize the model to a dictionary to be used by meilisearch.
    - meili_geo: Return the geo-location for the model. (If the model supports geolocation, else raise a ValueError.)

    Example:
    ```python
    from django.db import models
    from django_meili.models import IndexMixin

    class Post(IndexMixin, models.Model):
        title = models.CharField(max_length=255)
        body = models.TextField()

        class MeiliMeta:
            filterable_fields = ("title",)
            searchable_fields = ("id", "title", "body")
            displayed_fields = ("id", "title", "body")

        def __str__(self):
            return self.title
    ```
    """

    meilisearch: IndexQuerySet
    _meilisearch: _Meili

    class MeiliMeta:
        displayed_fields: Iterable[str] = None
        searchable_fields: Iterable[str] = None
        filterable_fields: Iterable[str] = None
        sortable_fields: Iterable[str] = None
        supports_geo: bool = False
        index_name: str = None
        primary_key: str = "pk"
        include_pk_in_search: bool = False

    def __init_subclass__(cls) -> None:
        index_name = getattr(cls.MeiliMeta, "index_name", cls.__name__)
        primary_key = getattr(cls.MeiliMeta, "primary_key", "pk")
        displayed_fields = getattr(cls.MeiliMeta, "displayed_fields", None)
        searchable_fields = getattr(cls.MeiliMeta, "searchable_fields", None)
        filterable_fields = getattr(cls.MeiliMeta, "filterable_fields", None)
        sortable_fields = getattr(cls.MeiliMeta, "sortable_fields", None)
        supports_geo = getattr(cls.MeiliMeta, "supports_geo", False)
        include_pk_in_search = getattr(cls.MeiliMeta, "include_pk_in_search", False)

        if supports_geo:
            filterable_fields = ("_geo",) + (filterable_fields or ())
            sortable_fields = ("_geo",) + (sortable_fields or ())

        if settings.MEILISEARCH.get("OFFLINE", False):
            cls._meilisearch = _Meili(
                primary_key=primary_key,
                index_name=index_name,
                displayed_fields=displayed_fields,
                searchable_fields=searchable_fields,
                filterable_fields=filterable_fields,
                sortable_fields=sortable_fields,
                supports_geo=supports_geo,
                include_pk_in_search=include_pk_in_search,
                tasks=[],
            )
        else:
            _client.create_index(index_name, primary_key).with_settings(
                index_name,
                displayed_fields,
                searchable_fields,
                filterable_fields,
                sortable_fields,
            )

        cls._meilisearch = _Meili(
            primary_key=primary_key,
            index_name=index_name,
            displayed_fields=displayed_fields,
            searchable_fields=searchable_fields,
            filterable_fields=filterable_fields,
            sortable_fields=sortable_fields,
            supports_geo=supports_geo,
            include_pk_in_search=include_pk_in_search,
            tasks=[task for task in _client.tasks],
        )
        _client.flush_tasks()

        cls.meilisearch = IndexQuerySet(cls)

    def meili_filter(self) -> bool:
        """
        A function to decide if the model should be added to meilisearch.

        For example, if a post model could be a draft and that draft shouldn't
        be in the search database, then this filter can make sure its not added.

        By default it just returns True.
        """

        return True

    def meili_serialize(self):
        """
        How to serialize the model to a dictionary to be used by meilisearch.

        By default uses django.core.serializers.serialize and json.loads
        """

        from json import loads

        from django.core.serializers import serialize

        serialized_model = loads(
            serialize(
                "json",
                [self],
                use_natural_foreign_keys=True,
                use_natural_primary_keys=True,
            )
        )[0]

        if getattr(self.MeiliMeta, "include_pk_in_search", False):
            serialized_model["fields"][self.MeiliMeta.primary_key] = (
                self._meta.get_field(self.MeiliMeta.primary_key).value_to_string(self)
            )
        return serialized_model["fields"]

    def meili_geo(self) -> MeiliGeo:
        """Return the geo-location for the model.

        If the model does not support geolocation, raise a ValueError.
        """

        raise ValueError("Model does not support geolocation")

    class Meta:
        abstract = True



"""
_settings.py
Ian Kollipara <ian.kollipara@cune.edu>

This file contains the settings for the MeiliSearch Django app.
"""

from dataclasses import dataclass
from typing import TypedDict


class DjangoMeiliSettings(TypedDict):
    """
    Settings for the MeiliSearch Django app.

    Attributes:
        https: Whether to use HTTPS or not.
        host: The host for the MeiliSearch instance.
        master_key: The master key for the MeiliSearch instance.
        port: The port for the MeiliSearch instance.
    """

    HTTPS: bool
    HOST: str
    MASTER_KEY: str
    PORT: int
    TIMEOUT: int | None
    CLIENT_AGENTS: tuple[str] | None
    DEBUG: bool | None
    SYNC: bool | None
    OFFLINE: bool | None
    DEFAULT_BATCH_SIZE: int = 1000


@dataclass(frozen=True, slots=True)
class _DjangoMeiliSettings:
    """
    Settings for the MeiliSearch Django app.

    Attributes:
        https: Whether to use HTTPS or not.
        host: The host for the MeiliSearch instance.
        master_key: The master key for the MeiliSearch instance.
        port: The port for the MeiliSearch instance.
    """

    https: bool
    host: str
    master_key: str
    port: int
    timeout: int | None
    client_agents: tuple[str] | None
    debug: bool
    sync: bool
    offline: bool
    batch_size: int

    @classmethod
    def from_settings(cls) -> "_DjangoMeiliSettings":
        from django.conf import settings

        return cls(
            https=settings.MEILISEARCH.get("HTTPS", False),
            host=settings.MEILISEARCH.get("HOST", "localhost"),
            master_key=settings.MEILISEARCH.get("MASTER_KEY", None),
            port=settings.MEILISEARCH.get("PORT", 7700),
            timeout=settings.MEILISEARCH.get("TIMEOUT", None),
            client_agents=settings.MEILISEARCH.get("CLIENT_AGENTS", None),
            debug=settings.MEILISEARCH.get("DEBUG", settings.DEBUG),
            sync=settings.MEILISEARCH.get("SYNC", False),
            offline=settings.MEILISEARCH.get("OFFLINE", False),
            batch_size=settings.MEILISEARCH.get("DEFAULT_BATCH_SIZE", 1000),
        )



# Generated by Django 5.1.3 on 2025-01-08 22:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('posts', '0004_nonstandardidpost'),
    ]

    operations = [
        migrations.CreateModel(
            name='UuidIdPost',
            fields=[
                ('id', models.UUIDField(primary_key=True, serialize=False)),
                ('title', models.CharField(max_length=255)),
                ('body', models.TextField()),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



"""
querysets.py
Ian Kollipara <ian.kollipara@gmail.com>

This module contains the QuerySet classes for the Django MeiliSearch app.
"""

# Imports
from typing import TYPE_CHECKING, Literal, NamedTuple, Self, Type

from django.db.models import Case, When

from ._client import client

if TYPE_CHECKING:
    from .models import IndexMixin


class Radius(NamedTuple):
    """A radius for a geosearch query."""

    lat: float | str
    lng: float | str
    radius: int


class BoundingBox(NamedTuple):
    """A bounding box for a geosearch query."""

    top_right: tuple[float | str, float | str]
    bottom_left: tuple[float | str, float | str]


class Point(NamedTuple):
    """A point for a geosearch query."""

    lat: float | str
    lng: float | str


class IndexQuerySet:
    """QuerySet for a MeiliSearch index.

    This class provides a way to interact with a MeiliSearch index for a given model.
    The queryset mimics the Django QuerySet API and provides methods to filter, sort, and search the index.
    """

    def __init__(self, model: Type["IndexMixin"]):
        self.model = model
        self.index = client.get_index(model._meilisearch["index_name"])
        self.__offset = 0
        self.__limit = 20
        self.__filters: list[str] = []
        self.__sort: list[str] = []
        self.__matching_strategy: Literal["last", "all"] = "last"
        self.__attributes_to_search_on: list[str] = ["*"]

    def __repr__(self):
        return f"<IndexQuerySet for {self.model.__name__}>"

    def __str__(self):
        return f"IndexQuerySet for {self.model.__name__}"

    def __getitem__(self, index):
        if isinstance(index, slice):
            self.__offset = index.start
            self.__limit = index.stop
            return self
        else:
            raise TypeError("IndexQuerySet indices must be slices")

    def count(self) -> int:
        """Returns the number of documents in the index.

        Note: This method is not specific to the current queryset and will return the total number of documents in the index.
        """

        return self.index.get_stats().number_of_documents

    def order_by(self, *fields: str):
        """Orders the queryset by the given fields.

        This mimics the Django QuerySet API and allows for ordering by multiple fields.
        The fields can be prefixed with "-" to indicate descending order.

        For example:
        ```python
        Model.meilisearch.order_by("field1", "-field2")
        ```

        For geosearch, the special geoPoint construct can be used.

        For example:
        ```python
        Model.meilisearch.order_by("geoPoint(lat, lng)")
        Model.meilisearch.order_by("-geoPoint(lat, lng)")
        ```
        """

        for field in fields:
            geopoint = "_" if "geoPoint" in field else ""
            if field.startswith("-"):
                self.__sort.append(f"{geopoint}{field[1:]}:desc")
            else:
                self.__sort.append(f"{geopoint}{field}:asc")
        return self

    def filter(self, *geo_filters, **filters) -> Self:
        """Filters the queryset by the given filters.

        This set of filtering mimics the Django QuerySet API and allows for filtering by multiple fields.
        The currently implemented filters are:
        - exact: Filters for an exact match.
        - gte: Filters for greater than or equal to.
        - gt: Filters for greater than.
        - lte: Filters for less than or equal to.
        - lt: Filters for less than.
        - in: Filters for a value in a list.
        - range: Filters for a value in a range.
        - exists: Filters for the existence of a field.
        - isnull: Filters for the nullness of a field.

        For example:
        ```python
        Model.meilisearch.filter(field1__exact="value", field2__gte=10, field3__in=[1, 2, 3])
        ```

        For geosearch, the Radius and BoundingBox classes can be used, and should be passed as unnamed arguments.
        If the model does not support geosearch, a TypeError will be raised.
        If the provided positional arguments are not of type Radius or BoundingBox, a TypeError will be raised.

        For example:
        ```python
        Model.meilisearch.filter(Radius(lat, lng, radius))
        Model.meilisearch.filter(BoundingBox(top_right, bottom_left))
        ```
        """

        for geo_filter in geo_filters:
            if not self.model._meilisearch["supports_geo"]:
                raise TypeError(
                    f"Model {self.model.__name__} does not support geo filters"
                )
            if not isinstance(geo_filter, (Radius, BoundingBox)):
                raise TypeError(
                    f"Unnamed Argument must be of type Radius or BoundingBox, not {type(geo_filter)}"
                )
            if isinstance(geo_filter, Radius):
                self.__filters.append(
                    f"_geoRadius({geo_filter.lat}, {geo_filter.lng}, {geo_filter.radius})"
                )
            elif isinstance(geo_filter, BoundingBox):
                self.__filters.append(
                    f"_geoBoundingBox([{geo_filter.top_right[0]}, {geo_filter.top_right[1]}], [{geo_filter.bottom_left[0]}, {geo_filter.bottom_left[1]}])"
                )
        for filter, value in filters.items():
            if "__" not in filter or "__exact" in filter:
                if (
                    value == ""
                    or (isinstance(value, list) and len(value) == 0)
                    or value == {}
                ):
                    self.__filters.append(f"{filter.split('__')[0]} IS EMPTY")
                elif value is None:
                    self.__filters.append(f"{filter.split('__')[0]} IS NULL")
                else:
                    self.__filters.append(
                        f"{filter.split('__')[0]} = '{value}'"
                        if isinstance(value, str)
                        else f"{filter.split('__')[0]} = {value}"
                    )
            elif "__gte" in filter:
                if not isinstance(value, (int, float)):
                    raise TypeError(f"Cannot compare {type(value)} with int or float")
                self.__filters.append(f"{filter.split('__')[0]} >= {value}")
            elif "__gt" in filter:
                if not isinstance(value, (int, float)):
                    raise TypeError(f"Cannot compare {type(value)} with int or float")
                self.__filters.append(f"{filter.split('__')[0]} > {value}")
            elif "__lte" in filter:
                if not isinstance(value, (int, float)):
                    raise TypeError(f"Cannot compare {type(value)} with int or float")
                self.__filters.append(f"{filter.split('__')[0]} <= {value}")
            elif "__lt" in filter:
                if not isinstance(value, (int, float)):
                    raise TypeError(f"Cannot compare {type(value)} with int or float")
                self.__filters.append(f"{filter.split('__')[0]} < {value}")
            elif "__in" in filter:
                if not isinstance(value, list):
                    raise TypeError(f"Cannot compare {type(value)} with list")
                self.__filters.append(f"{filter.split('__')[0]} IN {value}")
            elif "__range" in filter:
                if not isinstance(value, (range, list, tuple)):
                    raise TypeError(
                        f"Cannot compare {type(value)} with range, list or tuple"
                    )
                self.__filters.append(
                    f"{filter.split('__')[0]} {value[0]} TO {value[1]}"
                    if not isinstance(value, range)
                    else f"{filter.split('__')[0]} {value.start} TO {value.stop}"
                )
            elif "__exists" in filter:
                if not isinstance(value, bool):
                    raise TypeError(f"Cannot compare {type(value)} with bool")
                self.__filters.append(
                    f"{filter.split('__')[0]} {'NOT ' if not value else ''}EXISTS"
                )
            elif "__isnull" in filter:
                if not isinstance(value, bool):
                    raise TypeError(f"Cannot compare {type(value)} with bool")
                self.__filters.append(
                    f"{filter.split('__')[0]} {'NOT ' if not value else ''}IS NULL"
                )

        return self

    def matching_strategy(self, strategy: Literal["last", "all"]):
        """Sets the matching strategy for the search.

        The matching strategy can be either "last" or "all".
        """

        self.__matching_strategy = strategy
        return self

    def attributes_to_search_on(self, *attributes):
        """Sets the attributes to search on.

        This method allows for setting the attributes to search on for the search query.

        For example:
        ```python
        Model.meilisearch.attributes_to_search_on("title", "body")
        ```
        """

        self.__attributes_to_search_on.append(*attributes)
        return self

    def search(self, q: str = ""):
        """Searches the index for the given query.

        This method searches the index for the given query and returns the results as an actual Django QuerySet.

        For example:
        ```python
        Model.meilisearch.search("Hello World") # Returns a Django QuerySet
        ```
        """

        results = self.index.search(
            q,
            {
                "offset": self.__offset,
                "limit": self.__limit,
                "filter": self.__filters,
                "sort": self.__sort,
                "matchingStrategy": self.__matching_strategy,
                "attributesToSearchOn": self.__attributes_to_search_on,
            },
        )
        id_field = getattr(self.model.MeiliMeta, "primary_key", "id")
        pk_list = [hit[id_field] for hit in results.get("hits", [])]
        preserved_order = Case(
            *[When(pk=pk, then=pos) for pos, pk in enumerate(pk_list)]
        )
        return self.model.objects.filter(pk__in=pk_list).order_by(preserved_order)



"""
apps.py
Ian Kollipara <ian.kollipara@gmail.com>

This module contains the Django app configuration for the Django MeiliSearch app.
"""

from typing import TYPE_CHECKING, Iterable, TypedDict

from django.apps import AppConfig
from meilisearch.models.task import TaskInfo

if TYPE_CHECKING:
    from django_meili.models import IndexMixin


class DjangoMeiliConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_meili"

    def ready(self):
        """
        Register all IndexMixin subclasses with two signal methods,
        which will sync with meilisearch.
        """

        from django.conf import settings
        from django.db.models.signals import post_delete, post_save

        from ._client import client as _client
        from .models import IndexMixin

        def add_model(**kwargs):
            """Add a model to the MeiliSearch index.

            This function is called when a model is saved and adds the model to the MeiliSearch index.
            """

            model: IndexMixin = kwargs["instance"]
            if model.meili_filter():
                serialized = model.meili_serialize()

                # Since the primary key can be any field, we need to check if it is 'pk' or another field.
                # If its 'pk', we can just use the model.pk, otherwise we need to get the value from the field.
                pk = (
                    model._meta.pk.value_to_string(model)
                    if model._meilisearch["primary_key"] == "pk"
                    else model._meta.get_field(
                        model._meilisearch["primary_key"]
                    ).value_to_string(model)
                )

                # This bit makes sure that geo is only added if the model supports it.
                geo = model.meili_geo() if model._meilisearch["supports_geo"] else None
                if settings.MEILISEARCH.get("OFFLINE", False):
                    return
                task = _client.get_index(
                    model._meilisearch["index_name"]
                ).add_documents(
                    [
                        serialized
                        | {"id": pk, "pk": model._meta.pk.value_to_string(model)}
                        | ({"_geo": geo} if geo else {})
                    ]
                )
                if settings.DEBUG:
                    finished = _client.wait_for_task(task.task_uid)
                    if finished.status == "failed":
                        raise Exception(finished)

        def delete_model(**kwargs):
            """Delete a model from the MeiliSearch index.

            This function is called when a model is deleted and removes the model from the MeiliSearch index.
            """

            model: IndexMixin = kwargs["instance"]
            if model.meili_filter():
                # Since the primary key can be any field, we need to check if it is 'pk' or another field.
                # If its 'pk', we can just use the model.pk, otherwise we need to get the value from the field.
                pk = (
                    model._meta.get_field(
                        model._meilisearch["primary_key"]
                    ).value_from_object(model)
                    if model._meilisearch["primary_key"] != "pk"
                    else model.pk
                )

                if settings.MEILISEARCH.get("OFFLINE", False):
                    return
                task = _client.get_index(
                    model._meilisearch["index_name"]
                ).delete_document(pk)
                if settings.DEBUG:
                    finished = _client.wait_for_task(task.task_uid)
                    if finished.status == "failed":
                        raise Exception(finished)

        # This loop connects the add_model and delete_model functions to the post_save and post_delete signals of all
        # Its also why the `django_meili` app needs to be loaded before all the user apps in the `INSTALLED_APPS` list.
        for model in IndexMixin.__subclasses__():
            post_save.connect(add_model, sender=model)
            post_delete.connect(delete_model, sender=model)



"""
_client.py
Ian Kollipara <ian.kollipara@gmail.com>

This module contains the MeiliSearch client for the Django MeiliSearch app.
"""

from typing import Self

from meilisearch.client import Client as _Client
from meilisearch.models.task import Task
from meilisearch.task import TaskInfo

from ._settings import _DjangoMeiliSettings


class Client:
    """MeiliSearch client for Django MeiliSearch.

    This class is a wrapper around the MeiliSearch client that provides
    a more Django-like interface for interacting with the MeiliSearch
    server.
    """

    def __init__(self, settings: _DjangoMeiliSettings):
        self.client = _Client(
            f"http{'s' if settings.https else ''}://{settings.host}:{settings.port}",
            settings.master_key,
            timeout=settings.timeout,
            client_agents=settings.client_agents,
        )
        self.is_sync = settings.sync
        self.tasks = []

    def flush_tasks(self):
        """Flush all currently stored tasks."""

        self.tasks = []

    def with_settings(
        self,
        index_name: str,
        displayed_fields: list[str] | None = None,
        searchable_fields: list[str] | None = None,
        filterable_fields: list[str] | None = None,
        sortable_fields: list[str] | None = None,
    ):
        """Create a new index with the given settings.

        This method creates an index all at once with the desired settings.

        Args:
            index_name (str): The name of the index to create.
            primary_key (str): The primary key for the index.
            displayed_fields (list[str] | None): The fields to display in search results.
            searchable_fields (list[str] | None): The fields to search on.
            filterable_fields (list[str] | None): The fields to filter on.
            sortable_fields (list[str] | None): The fields to sort on.

        Returns:
            Self: The client object.
        """

        self.tasks.append(
            self._handle_sync(
                self.client.index(index_name).update_settings(
                    {
                        "displayedAttributes": displayed_fields or ["*"],
                        "searchableAttributes": searchable_fields or ["*"],
                        "filterableAttributes": filterable_fields or [],
                        "sortableAttributes": sortable_fields or [],
                    }
                )
            )
        )
        return self

    def create_index(self, index_name: str, primary_key: str):
        """Create a new index with the given name and primary key.

        Args:
            index_name (str): The name of the index to create.
            primary_key (str): The primary key for the index.

        Returns:
            dict: The response from the MeiliSearch server.
        """
        if index_name not in [i.uid for i in self.get_indexes()]:
            self.tasks.append(
                self._handle_sync(
                    self.client.create_index(index_name, {"primaryKey": primary_key})
                )
            )
        return self

    def get_index(self, index_name: str):
        """Get an index by name.

        Args:
            index_name (str): The name of the index to get.

        Returns:
            Index: The index with the given name.
        """

        return self.client.index(index_name)

    def wait_for_task(self, task_uid: str) -> Task | TaskInfo:
        """Wait for a task to finish.

        Args:
            task_uid (str): The UID of the task to wait for.

        Returns:
            Task | TaskInfo: The task object.
        """

        task = self.client.wait_for_task(task_uid)
        return self._handle_sync(task)

    def get_indexes(self):
        """Get all indexes.

        Returns:
            list[Index]: A list of all indexes.
        """

        return self.client.get_indexes()["results"]

    def update_display(self, index_name: str, attributes: dict | None) -> Self:
        if attributes is None:
            return self
        self._handle_sync(
            self.client.index(index_name).update_displayed_attributes(attributes)
        )
        return self

    def update_searchable(self, index_name: str, attributes: dict | None) -> Self:
        if attributes is None:
            return self
        self._handle_sync(
            self.client.index(index_name).update_searchable_attributes(attributes)
        )
        return self

    def update_filterable(self, index_name: str, attributes: dict | None) -> Self:
        if attributes is None:
            return self
        self._handle_sync(
            self.client.index(index_name).update_filterable_attributes(attributes)
        )
        return self

    def update_sortable(self, index_name: str, attributes: dict | None) -> Self:
        if attributes is None:
            return self
        self._handle_sync(
            self.client.index(index_name).update_sortable_attributes(attributes)
        )
        return self

    def _handle_sync(self, task: TaskInfo) -> Task | TaskInfo:
        """Handle the sync task."""

        if self.is_sync:
            task = self.client.wait_for_task(task.task_uid)
            if task.status == "failed":
                raise Exception(task.error)
        return task


client = Client(_DjangoMeiliSettings.from_settings())



# Generated by Django 4.2.6 on 2024-10-24 02:20

from django.db import migrations, models
import posts.models


class Migration(migrations.Migration):
    dependencies = [
        ("posts", "0003_postnogeo"),
    ]

    operations = [
        migrations.CreateModel(
            name="NonStandardIdPost",
            fields=[
                (
                    "crazy_id",
                    models.CharField(
                        default=posts.models.rand_id,
                        max_length=128,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("title", models.CharField(max_length=255)),
                ("body", models.TextField()),
            ],
            options={
                "verbose_name": "NonStandard Id Post",
                "verbose_name_plural": "NonStandard IdPosts",
            },
        ),
    ]



import random
import string
from uuid import uuid4

from django.db import models

from django_meili.models import IndexMixin, MeiliGeo

# Create your models here.


class PostNoGeo(IndexMixin, models.Model):
    """Model definition for Post."""

    title = models.CharField(max_length=255)
    body = models.TextField()

    class Meta:
        """Meta definition for Post."""

        verbose_name = "Post"
        verbose_name_plural = "Posts"

    class MeiliMeta:
        filterable_fields = ("title",)
        searchable_fields = ("id", "title", "body")
        displayed_fields = ("id", "title", "body")
        index_name = "posts_not_geo"

    def __str__(self):
        return self.title


class Post(IndexMixin, models.Model):
    """Model definition for Post."""

    title = models.CharField(max_length=255)
    body = models.TextField()
    lat = models.FloatField()
    lng = models.FloatField()

    def meili_geo(self) -> MeiliGeo:
        return {
            "lat": self.lat,
            "lng": self.lng,
        }

    class Meta:
        """Meta definition for Post."""

        verbose_name = "Post"
        verbose_name_plural = "Posts"

    class MeiliMeta:
        filterable_fields = ("title",)
        searchable_fields = ("id", "title", "body")
        displayed_fields = ("id", "title", "body")
        supports_geo = True

    def __str__(self):
        return self.title


def rand_id():
    return "".join(random.choices(string.ascii_letters, k=8))


class NonStandardIdPost(IndexMixin, models.Model):
    crazy_id = models.CharField(max_length=128, default=rand_id, primary_key=True)
    title = models.CharField(max_length=255)
    body = models.TextField()

    class Meta:
        """Meta definition for Post."""

        verbose_name = "NonStandard Id Post"
        verbose_name_plural = "NonStandard IdPosts"

    class MeiliMeta:
        primary_key = "crazy_id"  # test focus
        include_pk_in_search = True  # test focus
        filterable_fields = ("title",)
        searchable_fields = ("crazy_id", "title", "body")
        displayed_fields = ("crazy_id", "title", "body")

    def __str__(self):
        return self.title


class UuidIdPost(IndexMixin, models.Model):
    id = models.UUIDField(primary_key=True, default=uuid4)
    title = models.CharField(max_length=255)
    body = models.TextField()

    class MeiliMeta:
        filterable_fields = ("title",)
        searchable_fields = ("id", "title", "body")
        displayed_fields = ("id", "title", "body")

    def __str__(self):
        return self.title

class IndexNamePost(IndexMixin, models.Model):
    """Model definition for Post."""

    title = models.CharField(max_length=255)
    body = models.TextField()

    class Meta:
        """Meta definition for Post."""

        verbose_name = "Post"
        verbose_name_plural = "Posts"

    class MeiliMeta:
        filterable_fields = ("title",)
        searchable_fields = ("id", "title", "body")
        displayed_fields = ("id", "title", "body")
        index_name = "custom_index_name"

    def __str__(self):
        return self.title


# Generated by Django 4.2.6 on 2024-05-14 13:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('posts', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='post',
            name='lat',
            field=models.FloatField(default=0.0),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='post',
            name='lng',
            field=models.FloatField(default=0.0),
            preserve_default=False,
        ),
    ]



# Generated by Django 5.1.6 on 2025-02-25 21:36

import uuid
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('posts', '0005_uuididpost'),
    ]

    operations = [
        migrations.CreateModel(
            name='IndexNamePost',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=255)),
                ('body', models.TextField()),
            ],
            options={
                'verbose_name': 'Post',
                'verbose_name_plural': 'Posts',
            },
        ),
        migrations.AlterField(
            model_name='uuididpost',
            name='id',
            field=models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False),
        ),
    ]



"""
syncindex.py
Ian Kollipara <ian.kollipara@gmail.com>

This module contains the SyncIndexCommand class for the Django MeiliSearch app.
"""

from typing import TYPE_CHECKING

from django.apps import apps
from django.conf import settings
from django.core.management.base import BaseCommand

from django_meili._client import client as _client
from django_meili.models import IndexMixin

DEFAULT_BATCH_SIZE = settings.MEILISEARCH.get("DEFAULT_BATCH_SIZE", 1000)


def batch_qs(qs, batch_size=DEFAULT_BATCH_SIZE):
    """
    Returns a (start, end, total, queryset) tuple for each batch in the given
    queryset.

    Usage:
        # Make sure to order your querset
        article_qs = Article.objects.order_by('id')
        for start, end, total, qs in batch_qs(article_qs):
            print "Now processing %s - %s of %s" % (start + 1, end, total)
            for article in qs:
                print article.body
    """
    total = qs.count()
    for start in range(0, total, batch_size):
        end = min(start + batch_size, total)
        yield qs[start:end]


class Command(BaseCommand):
    help = "Syncs the MeiliSearch index for the given model."

    def add_arguments(self, parser):
        parser.add_argument(
            "model",
            type=str,
            help="The model to sync the index for. This should be in the format <app_name>.<model_name>",
        )
        parser.add_argument(
            "--batch_size",
            action="store_true",
            default=DEFAULT_BATCH_SIZE,
            help="The batch size you want to import in (default: 1000)",
        )

    def handle(self, *args, **options):
        Model = self._resolve_model(options["model"])
        tasks = []
        for qs in batch_qs(Model.objects.all(), options["batch_size"]):
            tasks.append(_client.get_index(Model._meilisearch["index_name"]).add_documents(
                [self._serialize(m) for m in qs if m.meili_filter()]
            ))
        for task in tasks:
            finished = _client.wait_for_task(task.task_uid)
            if finished.status == "failed":
                self.stderr.write(self.style.ERROR(finished.error))
                exit(1)
        self.stdout.write(self.style.SUCCESS(f"Synced index for {options['model']}"))

    def _serialize(self, model: IndexMixin) -> dict:
        """
        Serialize the model instance into a dictionary.
        """

        serialized = model.meili_serialize()
        pk = model._meta.pk.value_to_string(model)
        return serialized | {"id": pk, "pk": pk}

    def _resolve_model(self, model: str):
        """
        Resolve the model from the given string.
        """
        try:
            Model = apps.get_model(model)
            if IndexMixin not in Model.__mro__:
                raise ValueError("Model does not inherit from IndexMixin")
        except LookupError as e:
            self.stdout.write(self.style.ERROR(f"Model not found: {model}"))
            exit(1)
        except ValueError as e:
            self.stdout.write(self.style.ERROR(f"Invalid model: {model}"))
            exit(1)
        return Model



from random import uniform
from io import StringIO

from django.db import models
from django.test import TestCase, override_settings
from django.test.utils import isolate_apps
from django.core import management

from django_meili.models import IndexMixin, MeiliGeo
from django_meili.querysets import Radius
from posts.models import NonStandardIdPost, Post, PostNoGeo, UuidIdPost, IndexNamePost

# Create your tests here.


def generate_random_coordinates():
    # Define the range of latitude and longitude
    min_lat, max_lat = -90.0, 90.0  # Range for latitude (-90 to 90 degrees)
    min_lon, max_lon = -180.0, 180.0  # Range for longitude (-180 to 180 degrees)

    # Generate random coordinates within the defined range
    latitude = uniform(min_lat, max_lat)
    longitude = uniform(min_lon, max_lon)

    return latitude, longitude


@isolate_apps("posts", attr_name="apps")
@override_settings(MEILISEARCH={"OFFLINE": True}, DEBUG=True)
class OfflineDjangoMeiliTestCase(TestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        class PostNoGeo(IndexMixin, models.Model):
            """Model definition for Post."""

            title = models.CharField(max_length=255)
            body = models.TextField()

            class MeiliMeta:
                filterable_fields = ("title",)
                searchable_fields = ("id", "title", "body")
                displayed_fields = ("id", "title", "body")
                index_name = "posts_not_geo"

        class Post(IndexMixin, models.Model):
            """Model definition for Post."""

            title = models.CharField(max_length=255)
            body = models.TextField()
            lat = models.FloatField()
            lng = models.FloatField()

            def meili_geo(self) -> MeiliGeo:
                return {
                    "lat": self.lat,
                    "lng": self.lng,
                }

            class MeiliMeta:
                filterable_fields = ("title",)
                searchable_fields = ("id", "title", "body")
                displayed_fields = ("id", "title", "body")
                supports_geo = True

        cls.Post = Post
        cls.PostNoGeo = PostNoGeo
        return super().setUpTestData()

    def test_offline_is_set(self):
        from django.conf import settings

        self.assertTrue(settings.MEILISEARCH.get("OFFLINE", False))

    def test_offline_does_not_create_index(self):
        self.assertEqual(self.Post._meilisearch["tasks"], [])
        self.assertEqual(self.PostNoGeo._meilisearch["tasks"], [])


@override_settings(MEILISEARCH={"SYNC": True}, DEBUG=True)
class DjangoMeiliTestCase(TestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        cls.coordinates = generate_random_coordinates()
        cls.post = Post.objects.create(
            title="Hello World",
            body="This is a test post",
            lat=cls.coordinates[0],
            lng=cls.coordinates[1],
        )
        cls.post_no_geo = PostNoGeo.objects.create(
            title="Hello World", body="This is a test post"
        )

        return super().setUpTestData()

    @classmethod
    def tearDownClass(cls) -> None:
        from django_meili._client import client

        client.client.delete_index(Post._meilisearch["index_name"])
        client.client.delete_index(PostNoGeo._meilisearch["index_name"])
        return super().tearDownClass()

    def test_post_created(self):
        self.assertEqual(self.post.title, "Hello World")
        self.assertEqual(self.post.body, "This is a test post")

    def test_post_was_indexed(self):
        self.assertNotEqual(Post.meilisearch.count(), 0)

    def test_post_search_returns_post(self):
        self.assertEqual(
            Post.meilisearch.search("Hello World").first().title, "Hello World"
        )

    def test_bad_search_returns_nothing(self):
        self.assertEqual(Post.meilisearch.search("al;kdfja;lsdkfj").count(), 0)

    def test_post_search_can_be_filtered(self):
        self.assertEqual(
            Post.meilisearch.filter(title="Hello World").search().first().title,
            "Hello World",
        )

    def test_post_search_can_be_filtered_by_geo(self):
        self.assertEqual(
            Post.meilisearch.filter(
                Radius(self.coordinates[0], self.coordinates[1], 100)
            )
            .search()
            .first()
            .title,
            "Hello World",
        )

    def test_post_no_geo_throws_error_on_geo_filter(self):
        with self.assertRaises(TypeError):
            PostNoGeo.meilisearch.filter(Radius(0, 0, 100))

    def test_post_search_can_be_ordered_by_geo(self):
        self.assertEqual(
            Post.meilisearch.order_by(
                f"geoPoint({self.coordinates[0]}, {self.coordinates[1]})"
            )
            .search()
            .first()
            .title,
            "Hello World",
        )

    def test_post_no_geo_has_custom_index_name(self):
        self.assertEqual(PostNoGeo._meilisearch["index_name"], "posts_not_geo")

    def test_django_meili_only_makes_two_requests_per_index_creation(self):
        tasks = Post._meilisearch["tasks"]
        self.assertEqual(len(tasks), 2)
        tasks = PostNoGeo._meilisearch["tasks"]
        self.assertEqual(len(tasks), 2)

    def test_django_meili_does_not_sync_when_offline(self):
        post_no_geo_original_count = PostNoGeo.meilisearch.count()

        post2 = Post.objects.create(
            title="Hello World",
            body="This is a test post",
            lat=self.coordinates[0],
            lng=self.coordinates[1],
        )
        post_updated_count = Post.meilisearch.count()
        with override_settings(MEILISEARCH={"OFFLINE": True}):
            PostNoGeo.objects.create(
                title="Hello World",
                body="This is a test post",
            )
            post2.delete()

        self.assertEqual(PostNoGeo.meilisearch.count(), post_no_geo_original_count)
        self.assertEqual(Post.meilisearch.count(), post_updated_count)


@override_settings(MEILISEARCH={"SYNC": True}, DEBUG=True)
class DjangoMeiliNonStandardIdTestCase(TestCase):
    target_model = NonStandardIdPost

    @classmethod
    def setUpTestData(cls) -> None:
        cls.post = cls.target_model.objects.create(
            title="Hello World",
            body="This is a test post",
        )

        return super().setUpTestData()

    @classmethod
    def tearDownClass(cls) -> None:
        from django_meili._client import client

        client.client.delete_index(cls.target_model._meilisearch["index_name"])
        return super().tearDownClass()

    def test_post_created(self):
        self.assertEqual(self.post.title, "Hello World")
        self.assertEqual(self.post.body, "This is a test post")

    def test_post_was_indexed(self):
        self.assertNotEqual(self.target_model.meilisearch.count(), 0)

    def test_post_search_returns_post(self):
        self.assertEqual(
            self.target_model.meilisearch.search("Hello World").first().title,
            "Hello World",
        )

    def test_crazy_id_present_in_serializer(self):
        # {'title': 'Hello World', 'body': 'This is a test post', 'crazy_id': 'WqzyCvZF'}
        self.assertEqual(
            list(
                self.target_model.meilisearch.search("Hello World")
                .first()
                .meili_serialize()
                .keys()
            ),
            ["title", "body", "crazy_id"],
        )

    def test_bad_search_returns_nothing(self):
        self.assertEqual(
            self.target_model.meilisearch.search("al;kdfja;lsdkfj").count(), 0
        )

    def test_post_search_can_be_filtered(self):
        self.assertEqual(
            self.target_model.meilisearch.filter(title="Hello World")
            .search()
            .first()
            .title,
            "Hello World",
        )


@override_settings(MEILISEARCH={"SYNC": True}, DEBUG=True)
class DjangoMeiliUuidTestCase(TestCase):
    target_model = UuidIdPost

    @classmethod
    def setUpTestData(cls) -> None:
        cls.post = cls.target_model.objects.create(
            title="Hello World",
            body="This is a test post",
        )

        return super().setUpTestData()

    @classmethod
    def tearDownClass(cls) -> None:
        from django_meili._client import client

        client.client.delete_index(cls.target_model._meilisearch["index_name"])
        return super().tearDownClass()

    def test_post_created(self):
        self.assertEqual(self.post.title, "Hello World")
        self.assertEqual(self.post.body, "This is a test post")

    def test_post_was_indexed(self):
        self.assertNotEqual(self.target_model.meilisearch.count(), 0)

    def test_post_search_returns_post(self):
        self.assertEqual(
            self.target_model.meilisearch.search("Hello World").first().title,
            "Hello World",
        )

    def test_id_present_in_serializer(self):
        # {'title': 'Hello World', 'body': 'This is a test post', 'crazy_id': 'WqzyCvZF'}
        self.assertEqual(
            list(
                self.target_model.meilisearch.search("Hello World")
                .first()
                .meili_serialize()
                .keys()
            ),
            ["title", "body"],
        )

    def test_bad_search_returns_nothing(self):
        self.assertEqual(
            self.target_model.meilisearch.search("al;kdfja;lsdkfj").count(), 0
        )

    def test_post_search_can_be_filtered(self):
        self.assertEqual(
            self.target_model.meilisearch.filter(title="Hello World")
            .search()
            .first()
            .title,
            "Hello World",
        )

@override_settings(MEILISEARCH={"SYNC": True}, DEBUG=True)
class DjangoMeiliSyncindexCommandTestCase(TestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        cls.out = StringIO()

        return super().setUpTestData()

    @classmethod
    def tearDownClass(cls) -> None:
        from django_meili._client import client

        client.client.delete_index(IndexNamePost._meilisearch["index_name"])
        return super().tearDownClass()

    def test_syncindex_command(self):
        self.assertEqual(IndexNamePost._meilisearch["index_name"], "custom_index_name")
        # index is in sync
        IndexNamePost.objects.create(
            title="Hello World1", body="This is a test post1"
        )
        self.assertEqual(IndexNamePost.meilisearch.count(), 1)

        # index is out of sync
        with override_settings(MEILISEARCH={"OFFLINE": True}):
            IndexNamePost.objects.create(
                title="Hello World2", body="This is a test post2"
            )
        self.assertEqual(IndexNamePost.meilisearch.count(), 1)

        # index is in sync again
        management.call_command('syncindex', f'posts.{IndexNamePost.__name__}', stdout=self.out)
        self.assertIn("Synced index for posts.IndexNamePost", self.out.getvalue())
        self.assertEqual(IndexNamePost.meilisearch.count(), 2)

@override_settings(MEILISEARCH={"SYNC": True}, DEBUG=True)
class DjangoMeiliClearindexCommandTestCase(TestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        cls.out = StringIO()

        return super().setUpTestData()

    @classmethod
    def tearDownClass(cls) -> None:
        from django_meili._client import client

        client.client.delete_index(IndexNamePost._meilisearch["index_name"])
        return super().tearDownClass()

    def test_clearindex_command(self):
        self.assertEqual(IndexNamePost._meilisearch["index_name"], "custom_index_name")
        IndexNamePost.objects.create(
            title="Hello World", body="This is a test post"
        )
        self.assertEqual(IndexNamePost.meilisearch.count(), 1)

        management.call_command('clearindex', f'posts.{IndexNamePost.__name__}', stdout=self.out)
        self.assertIn("Cleared index for <class 'posts.models.IndexNamePost'>", self.out.getvalue())
        self.assertEqual(IndexNamePost.meilisearch.count(), 0)


"""Custom Django integration for MeiliSearch."""

__version__ = "0.0.12"

# Optional Support for DJP, a Django Plugin System
# https://djp.readthedocs.io/en/latest/index.html
try:
    import djp

    @djp.hookimpl
    def installed_apps():
        return ["django_meili"]

    @djp.hookimpl
    def settings(current_settings):
        current_settings["MEILISEARCH"] = current_settings.get("MEILISEARCH", {})

except (ImportError, ModuleNotFoundError):
    pass
